#include <httpd.h>
#include <http_protocol.h>
#include <http_config.h>
#include <syslog.h>
#include <zlib.h>

#define BUFFER 8192

#include "apr_general.h"
#include "apr_pools.h"
#include "apr_strings.h"
#include "apr_file_io.h"


/* mod_gzip - copyright 2007 ed neville
 * reads a .gz file to the client in substitution for the .html with
 * gzip content-encoding. if the client does not accept the gzip
 * encoding then a decompressed format will be sent.
 *
 * 2006-05-14 ed AT s5h DOT net - beta code
 * Use at your own risk. Author not responsible for use.
 *
 * Updates at http://www.s5h.net
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); 
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at 
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License. 
 *
 * $URL: svn+ssh://fw/repos/code/apache2_modules/gzip/mod_gzip_disk.c $
 * $Id: mod_gzip_disk.c 99 2008-02-13 19:05:06Z ed $
 */

// this function just outputs to the syslog a key/value pair.
void debug_print_chr_ptr( request_rec *r, const char *key, const char *val ) {
	char *ptr = (char *)apr_palloc( r->pool, 
		strlen( key ) + strlen( val ) + 3 );

	if( ptr != NULL ) {
		// sprintf( ptr, "%s: %s", key, val );
		syslog( LOG_DEBUG, "%s: %s", key, val );
	}
	else {
		syslog( LOG_DEBUG, "not enough room to create log" );
	}
}

// this is the callback function from the apr_table_do function. we're
// not using the generic function above since we want the prefix header:
int header_do( request_rec *r, const char *key, const char *val ) {
	char *temp = (char *)apr_palloc( r->pool, strlen( key ) + 20 );
	sprintf( temp, "Header %s", key );
	debug_print_chr_ptr( r, temp, val );
	return(1);
}

// this function determines if the user can accept the gzip encoding.
static int user_can_accept_gzip( request_rec *r ) {
	char *ptr;
	if( ptr = (void*)apr_table_get( r->headers_in, "Accept-Encoding" ) ) {
		if( strstr( ptr, "gzip" ) != NULL ) {
			return( OK );
		}
	}
	return( DECLINED );
}

// it would be nice to integrate with mod_mime here, but i'm not sure 
// how well that will go, must read the code a little more FIXME
static int set_mime_type( request_rec *r, char *s ) {
	char *ptr, *p;

	p = ptr = (char *)apr_palloc( r->pool, strlen( s ) + 1 );
	
	strncpy( ptr, s, strlen(s)+1 );

	while( strchr( ptr, '.' ) ) {
		ptr++;
	}

	if( ptr > p ) {
		(--ptr)[0] = '\0';
	}

	ptr = p;

	while( strchr( ptr, '.' ) ) {
		ptr++;
	}

	if( strncmp( ptr, "html", 4 ) == 0 
		|| strncmp( ptr, "php", 3 ) == 0 ) {
		ap_set_content_type( r, "text/html;charset=ascii" );
		return( EXIT_SUCCESS );
	}

	ap_set_content_type( r, "text/plain" );

	return( EXIT_SUCCESS );
}

// we read the headers from the saved headers file if possible while
// stripping any line breaks from the end of the lines
static int set_headers_from_file( apr_file_t *file, request_rec *r ) {
	char *s = apr_palloc( r->pool, BUFFER );
	char *ptr, *dup, *val;

	if( s == NULL ) {
		return( DECLINED );
	}

	while( apr_file_gets( s, BUFFER, file ) == OK ) {
		if( ( ptr = strchr( s, ':' ) ) != NULL )  {
			ptr[0] = '\0';
			ptr++;
			if( *ptr == ' ' ) {
				ptr++;
			}

			// remove any line breaks which may exist
			while( strchr( ptr, 0x0a ) || strchr( ptr, 0x0d ) ) {
				ptr[strlen(ptr)-1] = '\0';
			}

			dup = apr_pstrdup( r->pool, s );
			val = apr_pstrdup( r->pool, ptr );
			if( dup == NULL || val == NULL ) {
				return( DECLINED );
			}

			if( strstr( dup, "Content-Type" ) != NULL ) {
				ap_set_content_type( r, val );
				continue;
			}

			apr_table_setn( r->headers_out, dup, val );
		}
	}

	return( OK );
}

static int gzip_disk_header_parser( request_rec *r ) {

	struct stat s;
	apr_size_t content_read;
	// char *buffer = (void *)apr_palloc( r->pool, BUFFER );
	int size_of_scratch = strlen( r->filename ) + BUFFER;
	char *scratch = (char *)apr_palloc( r->pool, size_of_scratch ),
		*header_scratch = (char *)apr_palloc( r->pool, size_of_scratch + 10 );
	size_t count;
	gzFile *gz;
	apr_file_t *gz_file, *gz_headers;
	apr_status_t *file_read_status;

	// this is helpful for debugging purposes, but we don't need it
	// for day-to-day, business as usual work
	
	// we need at least BUFFER anyway, so don't shout about BUFFER
	// being greater than 4 (which is the most needed for the
	// extension .gz

	if( r->method_number != M_GET ) {
		return( DECLINED );
	}

	if( strlen( r->filename ) > 2 ) {
		if( strcmp( r->filename + ( strlen( r->filename ) - 3 ), ".gz" ) == 0 ) {
			// the user wanted the .gz content, so let's
			// give it 'em
			return( DECLINED );
		}
	}

	if( r->filename[strlen( r->filename ) -1 ] == '/' ) {
		snprintf( scratch, size_of_scratch, "%sindex.html.gz",
			r->filename );
	}
	else {
		// normal situation
		snprintf( scratch, size_of_scratch, "%s.gz", r->filename );
	}

	// now create header_scratch being the same path scratch but 
	// with the suffix of .header, so that we can create something
	// similar to the original request - if it exists.
	snprintf( header_scratch, size_of_scratch+10, "%s.header", scratch );

	// syslog( LOG_DEBUG, scratch );
	// now we're going to do things using apr rather than FILE
	
	if( apr_file_open( &gz_file, scratch, APR_READ, APR_OS_DEFAULT, r->pool ) != OK ) {
		// we cant find the file, lets leave it at that
		// syslog( LOG_DEBUG, "we cannot find the file" );
		return( DECLINED );
	}

	
	// now we need logic for the case where the user can accept
	// gzip, they have the appropriate accept header

	if( apr_file_open( &gz_headers, header_scratch, APR_READ, APR_OS_DEFAULT, r->pool ) != OK ) {
		set_mime_type( r, scratch );
	}
	else {
		if( set_headers_from_file( gz_headers, r ) != OK ) {
			return( DECLINED );
		}
	}

	if( user_can_accept_gzip( r ) == OK ) {
		apr_table_setn( r->headers_out, 
			"Content-Encoding", "gzip" );
		// we have sendfile?

		// set the size of content_read on entry, on exit we use it
		// to tell ap_rwrite how much data to send
		while( content_read != 0 ) {
			content_read = BUFFER;
			file_read_status = apr_file_read( gz_file, (void*)scratch, &content_read );

			if( file_read_status == OK ) {
				ap_rwrite( scratch, content_read, r );
			}

			if( file_read_status == APR_EOF ) {
				return( DONE );
			}

			if( file_read_status != OK ) {
				return( DECLINED );
			}
		}

		apr_file_close( gz_file );
		return( DONE );
	}
	else {
		gz = gzopen( scratch, "rb" );
		// syslog( LOG_DEBUG, "Going to do a decompressed read" );
		while( ( count = gzread( gz, scratch, BUFFER ) ) > 0 ) {
			ap_rwrite( scratch, count, r );
		}
		gzclose( gz );
		return( DONE );
	}

	return( DECLINED );
}

// this is the business for the hook, it's a content generator, so it's
// a default-style handler
static void gzip_disk_hooks( apr_pool_t *pool ) {
	ap_hook_header_parser( gzip_disk_header_parser, NULL, NULL, APR_HOOK_MIDDLE );
}

// this is the definition for the module
module AP_MODULE_DECLARE_DATA gzip_disk_module = {
	STANDARD20_MODULE_STUFF,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	gzip_disk_hooks
};

